<?php
namespace bhl_bind;
use Exception;

function bhl_twig(array $inc_path = [])
{
  array_unshift($inc_path, __DIR__ . "/../tpl/");
  $loader = new \Twig\Loader\FilesystemLoader($inc_path);

  $twig = new \Twig\Environment($loader, [
    'debug' => true, 
    'autoescape' => false, 
    'strict_variables' => true]
  );
  $twig->addExtension(new \Twig\Extension\DebugExtension());

  bhl_add_twig_support($twig);

  return $twig;
}

function bhl_add_twig_support(\Twig\Environment $twig)
{
  $twig->addTest(new \Twig\TwigTest('instanceof', 
    function($obj, $class) 
    { 
      return (new \ReflectionClass($class))->isInstance($obj);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('ns_last', 
    function($str) 
    { 
      return bhl_get_last_name($str);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('ns_prefix', 
    function($str) 
    { 
      return bhl_get_ns_prefix($str);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('bhl_type', 
    function($type) 
    { 
      return bhl_gen_type($type);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('bhl_type_ref', 
    function($type) 
    { 
      return bhl_type_ref(is_string($type) ? $type : bhl_gen_type($type));
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('native_type', 
    function($type) 
    { 
      //TODO: mtgCSCodegen instance is not really needed
      return bhl_gen_native_type(new \mtgCSCodegen(), $type);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('norm_name', 
    function($name) 
    { 
      return bhl_normalize_name($name);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('rtrim_comma', 
    function($str) 
    { 
      return rtrim(rtrim($str), ',');
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('replace_in', 
    function($txt, array $args = []) 
    { 
      return str_replace($args[1], $txt, $args[0]);
    },
    ['is_variadic' => true]
  ));
  $twig->addFunction(new \Twig\TwigFunction('Error', 
    function($e) 
    {
      throw new Exception($e);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('Warn', 
    function($e) 
    {
      echo "WARN: $e\n";
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('has_token', 
    function($o, $token) 
    { 
      return $o->hasToken($token);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('token', 
    function($o, $name) 
    {
      return $o->getToken($name);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('token_or', 
    function($o, $name, $v) 
    {
      if($o->hasToken($name))
        return $o->getToken($name);
      else
        return $v;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('count_required_args', 
    function(\mtgMetaFunc $fn) 
    {
      $req_args_num = 0;
      foreach($fn->getArgs() as $arg)
      {
        if($arg->hasToken('default'))
          break;
        ++$req_args_num;
      }
      return $req_args_num;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('count_default_args', 
    function(\mtgMetaFunc $fn) 
    {
      $def_args_num = 0;
      foreach($fn->getArgs() as $arg)
      {
        if($arg->hasToken('default'))
          ++$def_args_num;
        else if($def_args_num > 0)
          throw new Exception("Field '{$name}' in {$fn->getName()} is expected to be tagged with @default");
      }
      return $def_args_num;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('types_array', 
    function(\mtgType $type) 
    {
      $arr = array();
      if($type instanceof \mtgMultiType)
        $arr = $type->getValues();
      else
        $arr[] = $type;
      return $arr;
    }
  ));
}

function bhl_get_last_name($name)
{
  $items = explode('.', $name);
  return end($items);
}

function bhl_normalize_name($name)
{
  return str_replace('.', '__', $name);
}

function bhl_get_ns_prefix($name)
{
  $items = explode('.', $name);
  if(sizeof($items) == 1)
    return '';

  $prefix = '';
  for($i=0;$i<sizeof($items)-1;++$i)
    $prefix .= 'Nest("'.$items[$i].'").';

  return $prefix;
}

function bhl_type_ref($type_str)
{
  if(preg_match('~((\w|\.)+)\[\]~', $type_str, $ms))
    return "types.TArr(\"{$ms[1]}\")";
  else if(preg_match('~async\s+func\s+((\w|\.)+)\s*\(\)~', $type_str, $ms))
    return "types.TFunc(true, \"{$ms[1]}\")";
  else if(preg_match('~func\s+((\w|\.)+)\s*\(\)~', $type_str, $ms))
    return "types.TFunc(false, \"{$ms[1]}\")";
  else if(preg_match('~func\s+((\w|\.)+)\s*\(((\w|\.)+)\)~', $type_str, $ms))
    return "types.TFunc(false, \"{$ms[1]}\", \"{$ms[3]}\")";
  else if(preg_match('~((\w|\.)+)\,((\w|\.)+)~', $type_str, $ms))
    return "types.T(\"{$ms[1]}\", \"{$ms[3]}\")";
  else
    return "types.T(\"$type_str\")";
}

function bhl_gen_type(\mtgType $type = null)
{
  if($type === null)
    return "void";

  if($type instanceof \mtgBuiltinType)
  {
    if($type->isFloat() || $type->isDouble())
      return "float";
    else if($type->isInt() || $type->isUint())
      return "int";
  }
  else if($type instanceof \mtgMetaFunc)
  {
    $ret_type = $type->getReturnType() ? bhl_gen_type($type->getReturnType()) : 'void';

    $str = "";
    $str .= "func $ret_type(";
    foreach($type->getArgs() as $arg)
      $str .= bhl_gen_type($arg->getType()) . ",";
    $str = rtrim($str, ',');
    $str .= ")";
    return $str;
  }
  else if($type instanceof \mtgMultiType)
  {
    $str = '';
    foreach($type->getValues() as $sub_type)
      $str .= bhl_gen_type($sub_type) . ",";
    return rtrim($str, ",");
  }
  else if($type instanceof \mtgArrType)
    return bhl_gen_type($type->getValue()) . "[]";

  return ''.$type;
}

function bhl_gen_native_type(\mtgCSCodegen $gen, \mtgType $type)
{
  //any special case
  if($type == "any")
    return "bhl.Val";

  if($type instanceof \mtgArrType)
  {
    if($type->getValue()->getName() == "any")
      return "bhl.ValList";
    else
      return "HList<".bhl_gen_native_type($gen, $type->getValue()).">"; 
  }

  //func special case
  if($type instanceof \mtgMetaFunc)
    return "VM.FuncPtr";

  $alias = null;
  if($type instanceof \mtgUserType)
    $alias = $type->getToken('bhl_native_class');

  if($type instanceof \mtgArrType)
    $orig_name = $type->getValue()->getName();
  else
    $orig_name = $type->getName();
  if($alias)
    $type->setName($alias);

  $ntype = $gen->genNativeType($type);

  if($alias)
    $type->setName($orig_name);

  return $ntype;
}

